/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2005  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <string.h>

#include "Common.h"
#include "Vif.h"
#include "VU0.h"
#include "VUmicro.h"
#include "VifDma.h" 

// Extern variables
extern VIFregisters *_vifRegs;
extern vifStruct *_vif;

// Generic constants
static const unsigned int VIF0intc = 4;
static const unsigned int VIF1intc = 5;
static const unsigned int VIF0dmanum = 0;
static const unsigned int VIF1dmanum = 1;

static int cycles;

typedef void (*UNPACKFUNCTYPE)( u32 *dest, u32 *data );
typedef int  (*UNPACKPARTFUNCTYPE)( u32 *dest, u32 *data, int size );

typedef struct {
	UNPACKFUNCTYPE       funcU;
	UNPACKFUNCTYPE       funcS;
	UNPACKPARTFUNCTYPE   funcUpart;
	UNPACKPARTFUNCTYPE   funcSpart;

	int bsize;
	int dsize;
	int gsize;
	int qsize;
} VIFUnpackFuncTable;

/* block size; data size; group size; qword size; */
#define _UNPACK_TABLE32(name, bsize, dsize, gsize, qsize) \
   { UNPACK_##name,         UNPACK_##name, \
     UNPACK_##name##part,   UNPACK_##name##part, \
	 bsize, dsize, gsize, qsize },

#define _UNPACK_TABLE(name, bsize, dsize, gsize, qsize) \
   { UNPACK_##name##u,      UNPACK_##name##s, \
     UNPACK_##name##upart,  UNPACK_##name##spart, \
	 bsize, dsize, gsize, qsize },

// Main table for function unpacking
VIFUnpackFuncTable VIFfuncTable[16] = {
	_UNPACK_TABLE32(S_32, 12, 4, 4, 4)		// 0x0 - S-32
	_UNPACK_TABLE(S_16, 6, 2, 4, 2)			// 0x1 - S-16
	_UNPACK_TABLE(S_8, 3, 1, 4, 1)			// 0x2 - S-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0x3

	_UNPACK_TABLE32(V2_32, 24, 4, 1, 8)		// 0x4 - V2-32
	_UNPACK_TABLE(V2_16, 12, 2, 1, 4)		// 0x5 - V2-16
	_UNPACK_TABLE(V2_8, 6, 1, 1, 2)			// 0x6 - V2-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0x7
	
	_UNPACK_TABLE32(V3_32, 36, 4, 1, 12)	// 0x8 - V3-32
	_UNPACK_TABLE(V3_16, 18, 2, 1, 6)		// 0x9 - V3-16
	_UNPACK_TABLE(V3_8, 9, 1, 1, 3)			// 0xA - V3-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0xB

	_UNPACK_TABLE32(V4_32, 48, 4, 1, 16)	// 0xC - V4-32
	_UNPACK_TABLE(V4_16, 24, 2, 1, 8)		// 0xD - V4-16
	_UNPACK_TABLE(V4_8, 12, 1, 1, 4)		// 0xE - V4-8
	_UNPACK_TABLE32(V4_5, 6, 2, 1, 8)		// 0xF - V4-5
};

#if (defined(__i386__) || defined(__x86_64__)) && defined(__GNUC__)
// Main table for function unpacking
VIFUnpackFuncTable VIFfuncTableSSE[16] = {
	_UNPACK_TABLE32(S_32, 12, 4, 4, 4)		// 0x0 - S-32
	_UNPACK_TABLE(S_16, 6, 2, 4, 2)			// 0x1 - S-16
	_UNPACK_TABLE(S_8, 3, 1, 4, 1)			// 0x2 - S-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0x3

	_UNPACK_TABLE32(V2_32, 24, 4, 1, 8)		// 0x4 - V2-32
	_UNPACK_TABLE(V2_16, 12, 2, 1, 4)		// 0x5 - V2-16
	_UNPACK_TABLE(V2_8, 6, 1, 1, 2)			// 0x6 - V2-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0x7
	
	_UNPACK_TABLE32(V3_32, 36, 4, 1, 12)	// 0x8 - V3-32
	_UNPACK_TABLE(V3_16, 18, 2, 1, 6)		// 0x9 - V3-16
	_UNPACK_TABLE(V3_8, 9, 1, 1, 3)			// 0xA - V3-8
	{ NULL, NULL, NULL, NULL, 0, 0, 0, 0 },	// 0xB

	_UNPACK_TABLE32(V4_32, 48, 4, 1, 16)	// 0xC - V4-32
	_UNPACK_TABLE(V4_16, 24, 2, 1, 8)		// 0xD - V4-16
	_UNPACK_TABLE(V4_8, 12, 1, 1, 4)		// 0xE - V4-8
	_UNPACK_TABLE32(V4_5, 6, 2, 1, 8)		// 0xF - V4-5
};
#endif


void vif0FLUSH();
void vif1FLUSH();

void vifDmaInit() {
}

__inline static int _limit( int a, int max ) {
   return ( a > max ? max : a );
}

static void VIFunpack(u32 *data, vifCode *v, int size, const unsigned int VIFdmanum) {
	u32 *dest;
	unsigned int unpackType;
	UNPACKFUNCTYPE func;
	UNPACKPARTFUNCTYPE funcP;
	VIFUnpackFuncTable *ft;
	vifStruct *vif;
	VIFregisters *vifRegs;
	VURegs * VU;
	u8 *cdata = (u8*)data;
	int bytes;
	int memsize;

	if (VIFdmanum == 0) {
		VU = &VU0;
		vif = &vif0;
		vifRegs = vif0Regs;
		memsize = 0x1000;
	} else {
		VU = &VU1;
		vif = &vif1;
		vifRegs = vif1Regs;
		memsize = 0x4000;
	}

	v->addr &= memsize-1;
	dest = (u32*)(VU->Mem + v->addr);
  
#ifdef VIF_LOG
	VIF_LOG("VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
#endif

/*	if (vifRegs->cycle.cl > vifRegs->cycle.wl) {
	   SysPrintf( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
	}*/
	if (v->size != size) {
#ifdef VIF_LOG
		VIF_LOG("*PCSX2*: warning v->size != size\n");
#endif
	}
	if ((v->addr+size*4) > memsize) {
		SysPrintf("*PCSX2*: fixme unpack overflow\n");
		SysPrintf( "VIF%d UNPACK: Mode=%x, v->size=%d, size=%d, v->addr=%x\n", 
            VIFdmanum, v->cmd & 0xf, v->size, size, v->addr );
	}

	if (size == 0) {
		return;
	}

	_vifRegs = (VIFregisters*)vifRegs;
    _vif = vif;

	// The unpack type
	unpackType = v->cmd & 0xf;
#if (defined(__i386__) || defined(__x86_64__)) && defined(__GNUC__)
	if (cpucaps.hasStreamingSIMDExtensions) {
		if (vifRegs->mode == 0 && (v->cmd & 0x10) == 0) {
			ft = &VIFfuncTableSSE[ unpackType ];
		} else {
			ft = &VIFfuncTable[ unpackType ];
		}
	} else {
		ft = &VIFfuncTable[ unpackType ];
	}
#else
	ft = &VIFfuncTable[ unpackType ];
#endif

	// Assigning the normal upack function, the part type is assigned later
	func = vif->usn ? ft->funcU : ft->funcS;
	funcP = vif->usn ? ft->funcUpart : ft->funcSpart;

	if (func == NULL) {
		SysPrintf( "UNPACK %x, %x, UNHANDLED %x\n", v->addr, v->size, unpackType );
		return;
	}

	if(unpackType == 0xC && vifRegs->mode == 0 && vifRegs->cycle.cl == vifRegs->cycle.wl && !(vifRegs->code & 0x10000000)){
		memcpy((u8*)dest, cdata, size*4);
		size = 0;
		return;
	}

	
/*	{
		int i;
		for (i=0; i<size; i++) {
			VIF_LOG("[%d] %8.8x\n", i, data[i]);
		}
	}*/
	// Unpacking
	vif->wl = 0; vif->cl = 0;
	size*= 4;
	if (vifRegs->cycle.cl >= vifRegs->cycle.wl) { /* skipping write */
		while (size >= ft->qsize) {
			if (vif->cl < vifRegs->cycle.wl) { /* unpack one qword */
				bytes = funcP( dest, (u32*)cdata, 4 );
				cdata += bytes;
				size -= bytes;
				vif->wl++;
				if (vif->wl == vifRegs->cycle.wl) {
	   	      		vif->wl = 0;
				}
			}
			dest += 4;
			vif->cl++;
			if (vif->cl == vifRegs->cycle.cl) {
		    	vif->cl = 0;
			}
		}
#ifdef VIF_LOG
		VIF_LOG("remaining %d\n", size);
#endif
		if (size != 0) {
#ifdef VIF_LOG
			VIF_LOG("warning, end with size = %d\n", size);
#endif
		}
		while (size >= ft->dsize) {
			if (vif->cl < vifRegs->cycle.wl) { /* unpack one qword */
				size-= funcP(dest, (u32*)cdata, (size/ft->dsize)*ft->gsize);
				break;
			}
			dest += 4;
			vif->cl++;
			if (vif->cl == vifRegs->cycle.cl) {
		    	vif->cl = 0;
			}
		}
	} else
	if (vifRegs->cycle.cl < vifRegs->cycle.wl) { /* filling write */
#ifdef VIF_LOG
		VIF_LOG("*PCSX2*: filling write\n");
#endif
		while (size >= ft->bsize) {
			if (vif->wl == vifRegs->cycle.wl) {
         		vif->wl = 0;
			}
			func(dest, (u32*)cdata);
			if (vif->wl < vifRegs->cycle.cl) { /* unpack one qword */
				cdata += ft->bsize;
				size -= ft->bsize;
				vif->cl++;
				if (vif->cl == vifRegs->cycle.cl) {
	   	      		vif->cl = 0;
				}
			}
			dest += 12;
			vif->wl++;
		}
	} else {
		while ( size >= ft->bsize ) {
			func(dest, (u32*)cdata);
			dest += 12;
			cdata += ft->bsize;
			size -= ft->bsize;
			vif->cl++; vif->wl++;
			if (vif->cl == vifRegs->cycle.cl) {
				vif->cl = 0;
			}
			if (vif->wl == vifRegs->cycle.wl) {
				vif->wl = 0;
			}
		}
		while (size >= ft->dsize) {
			if (vif->cl < vifRegs->cycle.wl) { /* unpack one qword */
				funcP(dest, (u32*)cdata, (size/ft->dsize)*ft->gsize);
				break;
			}
			dest += 4;
			vif->cl++;
			if (vif->cl == vifRegs->cycle.cl) {
            	vif->cl = 0;
			}
		}
	}
	if (size > 0) {
#ifdef VIF_LOG
		VIF_LOG("*PCSX2*: warning size(%d) > 0 after unpack\n", size);
#endif
	}
}

static void vuExecMicro( u32 addr, const unsigned int VIFdmanum ) {
	VURegs * VU;
	void (*_vuExecMicro)();

	if (VIFdmanum == 0) {
		_vuExecMicro = Cpu->ExecuteVU0Block;
		VU = &VU0;
		vif0FLUSH();
	} else {
		_vuExecMicro = Cpu->ExecuteVU1Block;
		VU = &VU1;
		vif1FLUSH();
	}

	VU->vifRegs->itop = VU->vifRegs->itops;

	if (VIFdmanum == 1) {
		/* in case we're handling a VIF1 execMicro 
		   set the top with the tops value */
		VU->vifRegs->top = VU->vifRegs->tops;

		/* is DBF flag set in VIF_STAT? */
		if (VU->vifRegs->stat & 0x80) {
			/* it is, so set tops with base + ofst 
			   and clear stat DBF flag */
			VU->vifRegs->tops = VU->vifRegs->base;
			VU->vifRegs->stat &= ~0x80;
		} else {
			/* it is not, so set tops with base
			   and set the stat DBF flag */
			VU->vifRegs->tops = VU->vifRegs->base + VU->vifRegs->ofst;
			VU->vifRegs->stat |= 0x80;
		}
	}

	if (VIFdmanum == 0) {
		vu0ExecMicro(addr);
	} else {
		vu1ExecMicro(addr);
	}
}

void vif0Init() {
	vif0ch = (DMACh*)&psH[0x8000];
}

void vif0FLUSH() {
	while (VU0.VI[REG_VPU_STAT].UL & 0x1) {
		Cpu->ExecuteVU0Block();
	}
}

void vif0UNPACK(u32 *data) {
	int vifNum;
    int vl, vn;
    int len;

	vif0FLUSH();

    vl = (vif0.cmd     ) & 0x3;
    vn = (vif0.cmd >> 2) & 0x3;
    vif0.tag.addr = (data[0] & 0x3ff) * 16;
    vif0.usn = (data[0] >> 14) & 0x1;
    vifNum = (data[0] >> 16) & 0xff;
    if ( vifNum == 0 ) vifNum = 256;

    if ( vif0Regs->cycle.wl <= vif0Regs->cycle.cl ) {
        len = ((int)ceil( (double)((( 32 >> vl ) * ( vn + 1 )) * vifNum ) / 32 ));
    } else {
        int n = vif0Regs->cycle.cl * (vifNum / vif0Regs->cycle.wl) + 
                _limit( vifNum % vif0Regs->cycle.wl, vif0Regs->cycle.cl );
        len = ((int)ceil( (double)(( ( 32 >> vl ) * ( vn + 1 )) * n ) / 32 ));
    }
   
    vif0.tag.cmd  = vif0.cmd;
    vif0.tag.size = len;
    vif0Regs->offset = 0;
}

void _vif0mpgTransfer(u32 addr, u32 *data, int size) {
/*	SysPrintf("_vif0mpgTransfer addr=%x; size=%x\n", addr, size);
	{
		FILE *f = fopen("vu1.raw", "wb");
		fwrite(data, 1, size*4, f);
		fclose(f);
	}*/
	if (memcmp(VU0.Micro + addr, data, size * 4)) {
		memcpy(VU0.Micro + addr, data, size * 4);
		Cpu->ClearVU0(addr, size);
	}
}

int vif0transferData(u32 *data, int size) {
	int ret=0;

#ifdef VIF_LOG 
	VIF_LOG("VIFtransferData: cmd %x, size %x, vif0.tag.size %x\n", vif0.cmd, size, vif0.tag.size);
#endif
	if ((vif0.cmd & 0x60) == 0x60) { // UNPACK
#ifdef VIF_LOG 
		VIF_LOG("UNPACKData: cmd %x, size %x, vif0.tag.size %x\n", vif0.cmd, size, vif0.tag.size);
#endif
		if (size < vif0.tag.size) {
			VIFunpack(data, &vif0.tag, size, VIF0dmanum);
			cycles+= size / BIAS;
			vif0.tag.addr += size * 4;
			vif0.tag.size -= size; 
			ret = size;
		} else {
			VIFunpack(data, &vif0.tag, vif0.tag.size, VIF0dmanum);
			cycles+= vif0.tag.size / BIAS;
			ret = vif0.tag.size;
			vif0.tag.size = 0;
		}
	} else {
		switch (vif0.cmd) {
			case 0x20: // STMASK
				vif0Regs->mask = data[0];
#ifdef VIF_LOG
				VIF_LOG("STMASK == %x\n", vif0Regs->mask);
#endif
				ret = 1;
                vif0.tag.size = 0;
                break;

            case 0x30: // STROW
                switch (vif0.tag.addr) {
                    case 0: vif0Regs->r0 = data[0]; break;
                    case 1: vif0Regs->r1 = data[0]; break;
                    case 2: vif0Regs->r2 = data[0]; break;
                    case 3: vif0Regs->r3 = data[0]; break;
                }
				ret = 1;
                vif0.tag.addr++;
                vif0.tag.size--;
                break;

            case 0x31: // STCOL
                switch (vif0.tag.addr) {
                    case 0: vif0Regs->c0 = data[0]; break;
                    case 1: vif0Regs->c1 = data[0]; break;
                    case 2: vif0Regs->c2 = data[0]; break;
                    case 3: vif0Regs->c3 = data[0]; break;
                }
				ret = 1;
                vif0.tag.addr++;
                vif0.tag.size--;
                break;

            case 0x4A: // MPG
                if (size < vif0.tag.size) {
					_vif0mpgTransfer(vif0.tag.addr, data, size);
                    vif0.tag.addr += size * 4;
                    vif0.tag.size -= size; 
                    ret = size;
                } else {
					_vif0mpgTransfer(vif0.tag.addr, data, vif0.tag.size);
					ret = vif0.tag.size;
                    vif0.tag.size = 0;
                }
                break;
		}
    }
	if (vif0.tag.size <= 0) {
		vif0.cmd = 0;
	}

	return ret;
}

void vif0CMD(u32 *data, int size) {
	int vifNum;

	switch ( vif0.cmd & 0x7F ) {
		case 0x00: // NOP
        	vif0.cmd = 0;
            break;

        case 0x01: // STCYCL
            vif0Regs->cycle.cl =  data[0] & 0xff;
            vif0Regs->cycle.wl = (data[0] >> 8) & 0xff;
            vif0.cmd = 0;
            break;

        case 0x04: // ITOP
            vif0Regs->itops = data[0] & 0x3ff;
            vif0.cmd = 0;
            break;

        case 0x05: // STMOD
            vif0Regs->mode = data[0] & 0x3;
            vif0.cmd = 0;
            break;

        case 0x07: // MARK
            vif0Regs->mark = data[0] & 0xffff;
            vif0.cmd = 0;
            break;

        case 0x10: // FLUSHE
			vif0FLUSH();
            vif0.cmd = 0;
            break;

        case 0x14: // MSCAL
        case 0x15: // MSCALF
            vuExecMicro( ( data[0] & 0xffff ) * 8, VIF0dmanum );
            vif0.cmd = 0;
            break;

        case 0x17: // MSCNT
            vuExecMicro( -1, VIF0dmanum );
            vif0.cmd = 0;
            break;

        case 0x20: // STMASK
            break;

        case 0x30: // STROW
            vif0.tag.addr = 0;
            vif0.tag.size = 4;
            break;

        case 0x31: // STCOL
            vif0.tag.addr = 0;
            vif0.tag.size = 4;
            break;

        case 0x4A: // MPG
			vif0FLUSH();
            vifNum = (data[0] >> 16) & 0xff;
            if (vifNum == 0) vifNum = 256;
            vif0.tag.addr = ( data[0] & 0xffff ) * 8;
            vif0.tag.size = vifNum * 2;
            break;

        default:			
            vif0.cmd = 0;
            if ((vif0Regs->err & 0x6) == 0) {  //Mask Vifcode and DMA tag mismatch errors
				SysPrintf( "UNKNOWN VifCmd: %x\n", vif0.cmd );
            	vif0Regs->stat |= 1 << 13;
            }
            break;
	}
}

int VIF0transfer(u32 *data, int size) {
	int ret;

#ifdef VIF_LOG 
	VIF_LOG( "VIF0transfer: size %x (vif0.cmd %x)\n", size, vif0.cmd );
#endif

/*	{
		int i;
		for (i=0; i<size; i++) {
#ifdef VIF_LOG 
		   VIF_LOG( "[%d]: %x\n", i, data[i] );
#endif
		}
	}*/

	//vif0.irq = 0;
	while (size > 0) {
		if (vif0.cmd) {
			ret = vif0transferData(data, size);
			data+= ret; size-= ret;
			continue;
		}
		if(vif0.irq && vif0.cmd != 0x7) break;
		vif0.cmd = (data[0] >> 24);
		vif0Regs->code = data[0];

#ifdef VIF_LOG 
		VIF_LOG( "VIFtransfer: cmd %x, num %x, imm %x, size %x\n", vif0.cmd, (data[0] >> 16) & 0xff, data[0] & 0xffff, size );
#endif

		if ((vif0.cmd & 0x80) && !(vif0Regs->err & 0x1)) { //i bit of vifcode and not masked by VIF0_ERR
#ifdef VIF_LOG
			VIF_LOG( "Interrupt on VIFcmd: %x (INTC_MASK = %x)\n", vif0.cmd, psHu32(INTC_MASK) );
#endif
			vif0Regs->stat|= VIF0_STAT_INT;
			vif0.irq++;
		}

		if ((vif0.cmd & 0x60) == 0x60) {
			vif0UNPACK(data);
		} else {
			vif0CMD(data, size);
		}
		data++; 
		size--;
	}

	if (vif0.irq > 0) {
		return -2;
	}
	return 0;
}

int  _VIF0chain() {
	u32 *pMem;
	u32 qwc = vif0ch->qwc;
	u32 ret;

	if (vif0ch->qwc == 0) return 0;

	pMem = dmaGetAddr(vif0ch->madr);
	if (pMem == NULL) return -1;

	ret = VIF0transfer(pMem, vif0ch->qwc*4);
	vif0ch->madr+= qwc*16;
	vif0ch->qwc-= qwc;
	cycles+= (qwc * 4) / BIAS; /* guessing */
	return ret;
}

int _chainVIF0() {
	u32 *ptag;
	int id;
	int ret;
	int done=0;

	ptag = (u32*)dmaGetAddr(vif0ch->tadr); //Set memory pointer to TADR
	if (ptag == NULL) {					   //Is ptag empty?
		psHu32(DMAC_STAT)|= 1<<15;		   //If yes, set BEIS (BUSERR) in DMAC_STAT register
		return -1;						   //Return -1 as an error has occurred
	}

	id        = (ptag[0] >> 28) & 0x7;	   //ID for DmaChain copied from bit 28 of the tag
	vif0ch->qwc  = (u16)ptag[0];		   //QWC set to lower 16bits of the tag
	vif0ch->madr = ptag[1];				   //MADR = ADDR field

#ifdef VIF_LOG
	VIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, madr=%lx, tadr=%lx\n",
			ptag[1], ptag[0], vif0ch->qwc, id, vif0ch->madr, vif0ch->tadr);
#endif

	// Transfer dma tag if tte is set
	if (vif0ch->chcr & 0x40) {
		ret = VIF0transfer(ptag+2, 2);  //Transfer Tag
		if (ret == -1) return -1;
		if (ret == -2) done = 1;
		cycles++;
	}

	vif0ch->chcr = ( vif0ch->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );  //Transfer upper part of tag to CHCR bits 31-15

	done |= hwDmacSrcChainWithStack(vif0ch, id);
	ret = _VIF0chain();											   //Transfers the data set by the switch
	if (ret == -1) return -1;									   //There's been an error
	if (ret == -2) done = 1;									   //IRQ has been set by VifTransfer

	if ((vif0ch->chcr & 0x80) && (ptag[0] >> 31) && done == 0) {   //Check TIE bit of CHCR and IRQ bit of tag
#ifdef VIF_LOG
		VIF_LOG( "dmaIrq Set\n" );
#endif
		SysPrintf( "dmaIrq Set\n" );
		vif0Regs->stat|= VIF0_STAT_VIS;							   //Set the Tag Interrupt flag of VIF0_STAT
		return 1;												   //End Transfer
	}

	return done;												   //Return Done
}

int  vif0Interrupt() {
	int ret;

#ifdef VIF_LOG 
	VIF_LOG("vif0Interrupt: %8.8x\n", cpuRegs.cycle);
#endif

	if (vif0ch->chcr & 0x4 && vif0.done == 0) {
		cycles = 0;
		ret = _chainVIF0();
        if (ret != 0) vif0.done = 1;
        INT(0, cycles);
        return 0;
	}
	
	vif0ch->chcr &=~ 0x100;
	hwDmacIrq(DMAC_VIF0);
	vif0Regs->stat&= ~0xF000000; // FQC=0
	if (vif0.irq > 0) {
		vif0.irq--;
		hwIntcIrq(VIF0intc);
	}

	return 1;
}

void dmaVIF0() {
#ifdef VIF_LOG
	VIF_LOG("dmaVIF0 chcr = %lx, madr = %lx, qwc  = %lx\n"
			"        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
			vif0ch->chcr, vif0ch->madr, vif0ch->qwc,
			vif0ch->tadr, vif0ch->asr0, vif0ch->asr1 );
#endif
	if (vif0.irq > 0) {
		vif0.irq--;
		hwIntcIrq(VIF0intc);
		return;
	}

	cycles = 0;
	vif0Regs->stat|= 0x8000000; // FQC=8

	if (!(vif0ch->chcr & 0x4)) { // Normal Mode 
		_VIF0chain();
		INT(0, cycles);
		return;
	}

	if (_VIF0chain() != 0) {
		INT(0, cycles);
		return;
	}

	// Chain Mode
	vif0.done = 0;
	INT(0, cycles);
}

void vif0Write32(u32 mem, u32 value) {
	if (mem == 0x10003830) { // MARK
		vif0Regs->stat&= ~0x40;
		vif0Regs->mark = value;
	} else
	if (mem == 0x10003c10) { // FBRST
#ifdef VIF_LOG
		VIF_LOG("VIF0_FBRST write32 0x%8.8x\n", value);
#endif
		if (value & 0x1) {
			/* Reset VIF */
			vif0Reset();
			vif0Regs->stat&= ~0x0F000000; // FQC=0
			
		}
		if (value & 0x2) {
			/* Force Break the VIF */
			/* I guess we should stop the VIF dma here 
			   but not 100% sure (linuz) */
			vif0Regs->stat |= VIF0_STAT_VFS;
			SysPrintf("*PCSX2*: fixme VIF0 ForceBreak\n");
		}
		if (value & 0x4) {
			/* Stop VIF */
			/* Not completly sure about this, can't remember what game 
			   used this, but 'draining' the VIF helped it, instead of 
			   just stoppin the VIF (linuz) */
			
			vif0Regs->stat |= VIF0_STAT_VSS;
			//dmaVIF0();	// Drain the VIF  --- VIF Stops as not to outstrip dma source (refraction)
		}
		if (value & 0x8) {
			int cancel = 0;
			/* Cancel stall, first check if there is a stall to cancel, 
			   and then clear VIF1_STAT VSS|VFS|VIS|INT|ER0|ER1 bits */
			if ((vif0Regs->stat & (VIF0_STAT_VIS|VIF0_STAT_INT)) && vif0Regs->stat & (VIF0_STAT_VSS|VIF0_STAT_VIS|VIF0_STAT_VFS)) {
				cancel = 1;
			}
			vif0Regs->stat &= ~(VIF0_STAT_VSS | VIF0_STAT_VFS | VIF0_STAT_VIS |
								VIF0_STAT_INT | VIF0_STAT_ER0 | VIF0_STAT_ER1);
			if (cancel) {
				dmaVIF0();
			}
		}
	} else
	if (mem == 0x10003820) { // ERR
		vif0Regs->err = value;
	}
}

void vif0Reset() {
	/* Reset the whole VIF, meaning the internal pcsx2 vars
	   and all the registers */
	memset(&vif0, 0, sizeof(vif0));
	memset(vif0Regs, 0, sizeof(vif0Regs));
}

int  vif0Freeze(gzFile f, int Mode) {
	gzfreeze(&vif0, sizeof(vif0));

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


void vif1Init() {
	vif1ch = (DMACh*)&psH[0x9000];
}

void vif1FLUSH() {
	int _cycles;
	_cycles = VU1.cycle;
	while(VU0.VI[REG_VPU_STAT].UL & 0x100) {
		Cpu->ExecuteVU1Block();
	}
	cycles+= VU1.cycle - _cycles;
}

void vif1UNPACK(u32 *data) {
	int vifNum;
    int vl, vn;
    int len;

	vif1FLUSH();

    vl = (vif1.cmd     ) & 0x3;
    vn = (vif1.cmd >> 2) & 0x3;
    vif1.tag.addr = (data[0] & 0x3ff);
    vif1.usn = (data[0] >> 14) & 0x1;
    vifNum = (data[0] >> 16) & 0xff;
    if ( vifNum == 0 ) vifNum = 256;

    if ( vif1Regs->cycle.wl <= vif1Regs->cycle.cl ) {
        len = ((int)ceil( (double)((( 32 >> vl ) * ( vn + 1 )) * vifNum ) / 32 ));
    } else {
        int n = vif1Regs->cycle.cl * (vifNum / vif1Regs->cycle.wl) + 
                _limit( vifNum % vif1Regs->cycle.wl, vif1Regs->cycle.cl );
        len = ((int)ceil( (double)(( ( 32 >> vl ) * ( vn + 1 )) * n ) / 32 ));
    }
   if ( ( data[0] >> 15) & 0x1 ) {
        vif1.tag.addr += vif1Regs->tops;
    }    

    vif1.tag.addr *= 16;
    vif1.tag.cmd  = vif1.cmd;
    vif1.tag.size = len;
    vif1Regs->offset = 0;
}

void _vif1mpgTransfer(u32 addr, u32 *data, int size) {
/*	SysPrintf("_vif1mpgTransfer addr=%x; size=%x\n", addr, size);
	{
		FILE *f = fopen("vu1.raw", "wb");
		fwrite(data, 1, size*4, f);
		fclose(f);
	}*/
	if (memcmp(VU1.Micro + addr, data, size * 4)) {
		memcpy(VU1.Micro + addr, data, size * 4);
		Cpu->ClearVU1(addr, size);
	}
}

int vif1transferData(u32 *data, int size) {
	int ret=0;

#ifdef VIF_LOG 
	VIF_LOG("VIFtransferData: cmd %x, size %x, vif1.tag.size %x\n", vif1.cmd, size, vif1.tag.size);
#endif
	if ((vif1.cmd & 0x60) == 0x60) { // UNPACK
#ifdef VIF_LOG 
		VIF_LOG("UNPACKData: cmd %x, size %x, vif1.tag.size %x\n", vif1.cmd, size, vif1.tag.size);
#endif
		if (size < vif1.tag.size) {
			/* size is less that the total size, transfer is 
			   'in pieces' */
			VIFunpack(data, &vif1.tag, size, VIF1dmanum);
			cycles+= size / BIAS;
			vif1.tag.addr += size * 4;
			vif1.tag.size -= size; 
			ret = size;
		} else {
			/* we got all the data, transfer it fully */
			VIFunpack(data, &vif1.tag, vif1.tag.size, VIF1dmanum);
			cycles+= vif1.tag.size / BIAS;
			ret = vif1.tag.size;
			vif1.tag.size = 0;
		}
	} else {
		switch (vif1.cmd) {
			case 0x20: // STMASK
				vif1Regs->mask = data[0];
#ifdef VIF_LOG
				VIF_LOG("STMASK == %x\n", vif1Regs->mask);
#endif
				ret = 1;
                vif1.tag.size = 0;
                break;

            case 0x30: // STROW
                switch (vif1.tag.addr) {
                    case 0: vif1Regs->r0 = data[0]; break;
                    case 1: vif1Regs->r1 = data[0]; break;
                    case 2: vif1Regs->r2 = data[0]; break;
                    case 3: vif1Regs->r3 = data[0]; break;
                }
				ret = 1;
                vif1.tag.addr++;
                vif1.tag.size--;
                break;

            case 0x31: // STCOL
                switch (vif1.tag.addr) {
                    case 0: vif1Regs->c0 = data[0]; break;
                    case 1: vif1Regs->c1 = data[0]; break;
                    case 2: vif1Regs->c2 = data[0]; break;
                    case 3: vif1Regs->c3 = data[0]; break;
                }
				ret = 1;
                vif1.tag.addr++;
                vif1.tag.size--;
                break;

            case 0x4A: // MPG
                if (size < vif1.tag.size) {
					_vif1mpgTransfer(vif1.tag.addr, data, size);
                    vif1.tag.addr += size * 4;
                    vif1.tag.size -= size; 
                    ret = size;
                } else {
					_vif1mpgTransfer(vif1.tag.addr, data, vif1.tag.size);
					ret = vif1.tag.size;
                    vif1.tag.size = 0;
                }
                break;

            case 0x50: // DIRECT
            case 0x51: // DIRECTHL
                if (size < vif1.tag.size) {
                    GSgifTransfer2(data, size/4);
                    vif1.tag.size-= size;
					ret = size;
                } else {
                    GSgifTransfer2(data, vif1.tag.size/4);
					ret = vif1.tag.size;
                    vif1.tag.size = 0;
                }
				break;
		}
    }
	if (vif1.tag.size <= 0) {
		vif1.cmd = 0;
	}

	return ret;
}

void vif1CMD(u32 *data, int size) {
	int vifNum;
	int vifImm;

	switch ( vif1.cmd & 0x7F ) {
		case 0x00: // NOP
        	vif1.cmd = 0;
            break;

        case 0x01: // STCYCL
            vif1Regs->cycle.cl =  data[0] & 0xff;
            vif1Regs->cycle.wl = (data[0] >> 8) & 0xff;
            vif1.cmd = 0;
            break;

        case 0x02: // OFFSET
            vif1Regs->ofst  = data[0] & 0x3ff;
            vif1Regs->stat &= ~0x80;
            vif1Regs->tops  = vif1Regs->base;
            vif1.cmd = 0;
            break;

        case 0x03: // BASE
            vif1Regs->base = data[0] & 0x3ff;
            vif1.cmd = 0;
            break;

        case 0x04: // ITOP
            vif1Regs->itops = data[0] & 0x3ff;
            vif1.cmd = 0;
            break;

        case 0x05: // STMOD
            vif1Regs->mode = data[0] & 0x3;
            vif1.cmd = 0;
            break;

        case 0x06: // MSKPATH3
            vif1Regs->mskpath3 = (data[0] >> 15) & 0x1; 
            if ( vif1Regs->mskpath3 ) {
                psHu32(GIF_STAT) |= 0x2;
            } else {
                psHu32(GIF_STAT) &= ~0x2;
            }
            vif1.cmd = 0;
            break;

        case 0x07: // MARK
            vif1Regs->mark = data[0] & 0xffff;
            vif1.cmd = 0;
            break;

        case 0x10: // FLUSHE
        case 0x11: // FLUSH
        case 0x13: // FLUSHA
			vif1FLUSH();
            vif1.cmd = 0;
            break;

        case 0x14: // MSCAL
        case 0x15: // MSCALF
            vuExecMicro( ( data[0] & 0xffff ) * 8, VIF1dmanum );
            vif1.cmd = 0;
            break;

        case 0x17: // MSCNT
            vuExecMicro( -1, VIF1dmanum );
            vif1.cmd = 0;
            break;

        case 0x20: // STMASK
            break;

        case 0x30: // STROW
            vif1.tag.addr = 0;
            vif1.tag.size = 4;
            break;

        case 0x31: // STCOL
            vif1.tag.addr = 0;
            vif1.tag.size = 4;
            break;

        case 0x4A: // MPG
			vif1FLUSH();
            vifNum = (data[0] >> 16) & 0xff;
            if (vifNum == 0) vifNum = 256;
            vif1.tag.addr = ( data[0] & 0xffff ) * 8;
            vif1.tag.size = vifNum * 2;
            break;

        case 0x50: // DIRECT
        case 0x51: // DIRECTHL
			vifImm = data[0] & 0xffff;
            if (vifImm == 0) {
				vif1.tag.size = 65536 * 4;
			} else {
				vif1.tag.size = vifImm * 4;
			}
            break;

        default:			
            vif1.cmd = 0;
            if ((vif1Regs->err & 0x6) == 0) {  //Ignore vifcode and tag mismatch error
				SysPrintf( "UNKNOWN VifCmd: %x\n", vif1.cmd );
            	vif1Regs->stat |= 1 << 13;
            }
            break;
	}
}


int VIF1transfer(u32 *data, int size) {
	int ret;

#ifdef VIF_LOG 
	VIF_LOG( "VIF1transfer: size %x (vif1.cmd %x)\n", size, vif1.cmd );
#endif

/*	{
		int i;
		for (i=0; i<size; i++) {
#ifdef VIF_LOG 
		   VIF_LOG( "[%d]: %x\n", i, data[i] );
#endif
		}
	}*/

	//vif1.irq = 0;
	
	while (size > 0) {
		vif1Regs->stat &= ~VIF1_STAT_VPS_W;
		if (vif1.cmd) {			
			vif1Regs->stat |= VIF1_STAT_VPS_T;
			ret = vif1transferData(data, size);
			data+= ret; size-= ret;
			vif1Regs->stat &= ~VIF1_STAT_VPS_T;
			continue;
		}

		if(vif1.irq && vif1.cmd != 0x7) break;

		vif1.cmd = (data[0] >> 24);
		vif1Regs->code = data[0];

#ifdef VIF_LOG 
		VIF_LOG( "VIFtransfer: cmd %x, num %x, imm %x, size %x\n", vif1.cmd, (data[0] >> 16) & 0xff, data[0] & 0xffff, size );
#endif

		if ((vif1.cmd & 0x80) && !(vif1Regs->err & 0x1)) { //i bit on vifcode and not masked by VIF1_ERR
#ifdef VIF_LOG
			VIF_LOG( "Interrupt on VIFcmd: %x (INTC_MASK = %x)\n", vif1.cmd, psHu32(INTC_MASK) );
#endif
			vif1Regs->stat|= VIF1_STAT_INT;
			vif1Regs->stat|= VIF1_STAT_VIS;
			vif1.irq++;
		}

		vif1Regs->stat |= VIF1_STAT_VPS_D;
		if ((vif1.cmd & 0x60) == 0x60) {
			vif1UNPACK(data);
		} else {
			vif1CMD(data, size);
		}
		vif1Regs->stat &= ~VIF1_STAT_VPS_D;
		if(vif1.tag.size > 0) vif1Regs->stat |= VIF1_STAT_VPS_W;
		data++; 
		size--;
	}

	if (vif1.irq > 0) {
		return -2;
	}
	return 0;
}

int  _VIF1chain() {
	u32 *pMem;
	u32 qwc = vif1ch->qwc;
	u32 ret;

	if (vif1ch->qwc == 0) return 0;

	pMem = dmaGetAddr(vif1ch->madr);
	if (pMem == NULL) return -1;

	ret = VIF1transfer(pMem, vif1ch->qwc*4);
	vif1ch->madr+= qwc*16;
	vif1ch->qwc-= qwc;
	cycles+= (qwc*4) / BIAS ; /* guessing */
	return ret;
}

int _chainVIF1() {
	u32 *ptag;
	int id;
	int done=0;
	int ret;

	ptag = (u32*)dmaGetAddr(vif1ch->tadr); //Set memory pointer to TADR
	if (ptag == NULL) {                    //Is ptag empty?
		psHu32(DMAC_STAT)|= 1<<15;         //If yes, set BEIS (BUSERR) in DMAC_STAT register
		return -1;						   //Return -1 as an error has occurred
	}

	id        = (ptag[0] >> 28) & 0x7; //ID for DmaChain copied from bit 28 of the tag
	vif1ch->qwc  = (u16)ptag[0];       //QWC set to lower 16bits of the tag
	vif1ch->madr = ptag[1];            //MADR = ADDR field

#ifdef VIF_LOG
	VIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, madr=%lx, tadr=%lx\n",
			ptag[1], ptag[0], vif1ch->qwc, id, vif1ch->madr, vif1ch->tadr);
#endif

	// Transfer dma tag if tte is set
	if (vif1ch->chcr & 0x40) {
		ret = VIF1transfer(ptag+2, 2);  //Transfer Tag
		if (ret == -1) return -1;       //There has been an error
		if (ret == -2) done = 1;        //IRQ set by VIFTransfer
		cycles++;
	}

	vif1ch->chcr = ( vif1ch->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 ); //Transfer upper part of tag to CHCR bits 31-15

	done |= hwDmacSrcChainWithStack(vif1ch, id);
	ret = _VIF1chain();											   //Transfers the data set by the switch
	if (ret == -1) return -1;									   //There's been an error
	if (ret == -2) done = 1;									   //IRQ has been set by VifTransfer

	if ((vif1ch->chcr & 0x80) && (ptag[0] >> 31)) {			       //Check TIE bit of CHCR and IRQ bit of tag
#ifdef VIF_LOG
		VIF_LOG( "dmaIrq Set\n" );
#endif
		//SysPrintf( "VIF1dmaIrq Set\n" );
		vif1Regs->stat|= VIF1_STAT_VIS;							   //Set the Tag Interrupt flag of VIF1_STAT
		return 1;												   //End Transfer
	}
	return done;												   //Return Done
}

int _vif1Interrupt() {
	int ret;

#ifdef VIF_LOG 
	VIF_LOG("vif1Interrupt: %8.8x\n", cpuRegs.cycle);
#endif

	if (vif1ch->chcr & 0x4 && vif1.done == 0) {
		cycles = 0;
		ret = _chainVIF1();
		if (ret != 0) vif1.done = 1;
		INT(1, cycles);
		return 0;
	}
	vif1ch->chcr &=~ 0x100;
	hwDmacIrq(DMAC_VIF1);
	vif1Regs->stat&= ~0x1F000000; // FQC=0
	if (vif1.irq > 0) {
		vif1.irq--;
		hwIntcIrq(VIF1intc);
	}

	return 1;
}

int  vif1Interrupt() {
#if (defined(__i386__) || defined(__x86_64__)) && defined(CPU_LOG)
	u64 ticko = GetCPUTick();
#endif
	int ret;

	ret = _vif1Interrupt();

#if (defined(__i386__) || defined(__x86_64__)) && defined(CPU_LOG)
	stats.vif1count+= GetCPUTick() - ticko;
#endif
	return ret;
}

void _dmaVIF1() {
#ifdef VIF_LOG
	VIF_LOG("dmaVIF1 chcr = %lx, madr = %lx, qwc  = %lx\n"
			"        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
			vif1ch->chcr, vif1ch->madr, vif1ch->qwc,
			vif1ch->tadr, vif1ch->asr0, vif1ch->asr1 );
#endif

	/* Check if there is a pending irq */
	if (vif1.irq > 0) {
		vif1.irq--;
		hwIntcIrq(VIF1intc);
		return;
	}

	if (((psHu32(DMAC_CTRL) & 0xC) == 0x8)) { // VIF MFIFO
		return;
	}

	cycles = 0;
	vif1Regs->stat|= 0x10000000; // FQC=16

	if (!(vif1ch->chcr & 0x4)) { // Normal Mode 
		if ((vif1ch->chcr & 0x1)) { // to Memory
			_VIF1chain();
		} else { // from Memory
			u64 *pMem;
			int size;

			pMem = (u64*)dmaGetAddr(vif1ch->madr);
			if (pMem == NULL) {
				psHu32(DMAC_STAT)|= 1<<15;
				return;
			}
			for (size=vif1ch->qwc; size>0; size--) {
				if (size > 1) GSreadFIFO((u64*)&psH[0x5000]);
				pMem[0] = psHu64(0x5000);
				pMem[1] = psHu64(0x5008); pMem+= 2;
			}
			vif1Regs->stat&= ~0x1f000000;
			vif1ch->qwc = 0;
		}

		INT(1, cycles);
		return;
	}

	if (_VIF1chain() != 0) {
		INT(1, cycles);
		return;
	}
	
	// Chain Mode
	vif1.done = 0;
	INT(1, cycles);

}

void dmaVIF1() {
#if (defined(__i386__) || defined(__x86_64__)) && defined(CPU_LOG)
	u64 ticko = GetCPUTick();
#endif

	_dmaVIF1();
#if (defined(__i386__) || defined(__x86_64__)) && defined(CPU_LOG)
	stats.vif1count+= GetCPUTick() - ticko;
#endif
}

void vif1Write32(u32 mem, u32 value) {
	if (mem == 0x10003c30) { // MARK
#ifdef VIF_LOG
		VIF_LOG("VIF1_MARK write32 0x%8.8x\n", value);
#endif
		/* Clear mark flag in VIF1_STAT and set mark with 'value' */
		vif1Regs->stat&= ~VIF1_STAT_MRK;
		vif1Regs->mark = value;
	} else
	if (mem == 0x10003c10) { // FBRST
#ifdef VIF_LOG
		VIF_LOG("VIF1_FBRST write32 0x%8.8x\n", value);
#endif
		if (value & 0x1) {
			/* Reset VIF */
			vif1Reset();
			vif1Regs->stat&= ~0x1F000000; // FQC=0
			
		}
		if (value & 0x2) {
			/* Force Break the VIF */
			/* I guess we should stop the VIF dma here 
			   but not 100% sure (linuz) */
			vif1Regs->stat |= VIF1_STAT_VFS;
			SysPrintf("*PCSX2*: fixme VIF1 ForceBreak\n");			
		}
		if (value & 0x4) {
			/* Stop VIF */
			/* Not completly sure about this, can't remember what game 
			   used this, but 'draining' the VIF helped it, instead of 
			   just stoppin the VIF (linuz) */
			vif1Regs->stat |= VIF1_STAT_VSS;
			//dmaVIF1();	// Drain the VIF  --- VIF Stops as not to outstrip dma source (refraction)
		}
		if (value & 0x8) {
			int cancel = 0;
			/* Cancel stall, first check if there is a stall to cancel, 
			   and then clear VIF1_STAT VSS|VFS|VIS|INT|ER0|ER1 bits */
			if ((vif1Regs->stat & VIF1_STAT_INT) && vif1Regs->stat & (VIF1_STAT_VSS|VIF1_STAT_VIS|VIF1_STAT_VFS)) {
				cancel = 1;
			}

			vif1Regs->stat &= ~(VIF1_STAT_VSS | VIF1_STAT_VFS | VIF1_STAT_VIS |
								VIF1_STAT_INT | VIF1_STAT_ER0 | VIF1_STAT_ER1);
			if (cancel) {
				dmaVIF1();
			}
		}
	} else
	if (mem == 0x10003c20) { // ERR
#ifdef VIF_LOG
		VIF_LOG("VIF1_ERR write32 0x%8.8x\n", value);
#endif
		/* Set VIF1_ERR with 'value' */
		vif1Regs->err = value;
	} else
	if (mem == 0x10003c00) { // STAT
#ifdef VIF_LOG
		VIF_LOG("VIF1_STAT write32 0x%8.8x\n", value);
#endif
		/* Only FDR bit is writable, so mask the rest */
		vif1Regs->stat = (vif1Regs->stat & ~VIF1_STAT_FDR) | (value & VIF1_STAT_FDR);
		if (vif1Regs->stat & VIF1_STAT_FDR) {
			vif1Regs->stat|= 0x01000000;
		} else {
			vif1Regs->stat&= ~0x1F000000; // FQC=0
		}
	}

	/* Other registers are read-only so do nothing for them */
}

void vif1Reset() {
	/* Reset the whole VIF, meaning the internal pcsx2 vars
	   and all the registers */
	memset(&vif1, 0, sizeof(vif1));
	memset(vif1Regs, 0, sizeof(vif1Regs));
}

int  vif1Freeze(gzFile f, int Mode) {
	gzfreeze(&vif1, sizeof(vif1));

	return 0;
}




